@@ -1,5 +1,5 @@
 /* NBD client library in userspace.
- * Copyright (C) 2020 Red Hat Inc.
+ * Copyright (C) 2020-2022 Red Hat Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -28,13 +28,15 @@
 #include <errno.h>
 #include <assert.h>
 #include <sys/stat.h>
+#include <inttypes.h>
 
 #include <pthread.h>
 
 #include <libnbd.h>
 
-#include "rounding.h"
 #include "iszero.h"
+#include "minmax.h"
+#include "rounding.h"
 
 #include "nbdcopy.h"
 
@@ -50,13 +52,13 @@ get_next_offset (uint64_t *offset, uint64_t *count)
   bool r = false;               /* returning false means no more work */
 
   pthread_mutex_lock (&lock);
-  if (next_offset < src.size) {
+  if (next_offset < src->size) {
     *offset = next_offset;
 
     /* Work out how large this range is.  The last range may be
      * smaller than THREAD_WORK_SIZE.
      */
-    *count = src.size - *offset;
+    *count = src->size - *offset;
     if (*count > THREAD_WORK_SIZE)
       *count = THREAD_WORK_SIZE;
 
@@ -69,7 +71,7 @@ get_next_offset (uint64_t *offset, uint64_t *count)
      * are called from threads and not necessarily in monotonic order
      * so the progress bar would move erratically.
      */
-    progress_bar (*offset, dst.size);
+    progress_bar (*offset, src->size);
   }
   pthread_mutex_unlock (&lock);
   return r;
@@ -89,11 +91,13 @@ multi_thread_copying (void)
    */
   assert (threads > 0);
   assert (threads == connections);
+/*
   if (src.ops == &nbd_ops)
     assert (src.u.nbd.handles.size == connections);
   if (dst.ops == &nbd_ops)
     assert (dst.u.nbd.handles.size == connections);
-  assert (src.size != -1);
+*/
+  assert (src->size != -1);
 
   workers = malloc (sizeof (pthread_t) * threads);
   if (workers == NULL) {
@@ -147,9 +151,9 @@ worker_thread (void *indexp)
 
     assert (0 < count && count <= THREAD_WORK_SIZE);
     if (extents)
-      src.ops->get_extents (&src, index, offset, count, &exts);
+      src->ops->get_extents (src, index, offset, count, &exts);
     else
-      default_get_extents (&src, index, offset, count, &exts);
+      default_get_extents (src, index, offset, count, &exts);
 
     for (i = 0; i < exts.size; ++i) {
       struct command *command;
@@ -157,9 +161,9 @@ worker_thread (void *indexp)
       char *data;
       size_t len;
 
-      if (exts.ptr[i].hole) {
-        /* The source is a hole so we can proceed directly to
-         * skipping, trimming or writing zeroes at the destination.
+      if (exts.ptr[i].zero) {
+        /* The source is zero so we can proceed directly to skipping,
+         * fast zeroing, or writing zeroes at the destination.
          */
         command = calloc (1, sizeof *command);
         if (command == NULL) {
@@ -180,8 +184,8 @@ worker_thread (void *indexp)
          */
         while (exts.ptr[i].length > 0) {
           len = exts.ptr[i].length;
-          if (len > MAX_REQUEST_SIZE)
-            len = MAX_REQUEST_SIZE;
+          if (len > request_size)
+            len = request_size;
           data = malloc (len);
           if (data == NULL) {
             perror ("malloc");
@@ -208,11 +212,11 @@ worker_thread (void *indexp)
           wait_for_request_slots (index);
 
           /* Begin the asynch read operation. */
-          src.ops->asynch_read (&src, command,
-                                (nbd_completion_callback) {
-                                  .callback = finished_read,
-                                  .user_data = command,
-                                });
+          src->ops->asynch_read (src, command,
+                                 (nbd_completion_callback) {
+                                   .callback = finished_read,
+                                   .user_data = command,
+                                 });
 
           exts.ptr[i].offset += len;
           exts.ptr[i].length -= len;
@@ -254,7 +258,7 @@ wait_for_request_slots (uintptr_t index)
 static unsigned
 in_flight (uintptr_t index)
 {
-  return src.ops->in_flight (&src, index) + dst.ops->in_flight (&dst, index);
+  return src->ops->in_flight (src, index) + dst->ops->in_flight (dst, index);
 }
 
 /* Poll (optional) NBD src and NBD dst, moving the state machine(s)
@@ -271,7 +275,7 @@ poll_both_ends (uintptr_t index)
   /* Note: if polling is not supported, this function will
    * set fd == -1 which poll ignores.
    */
-  src.ops->get_polling_fd (&src, index, &fds[0].fd, &direction);
+  src->ops->get_polling_fd (src, index, &fds[0].fd, &direction);
   if (fds[0].fd >= 0) {
     switch (direction) {
     case LIBNBD_AIO_DIRECTION_READ:
@@ -286,7 +290,7 @@ poll_both_ends (uintptr_t index)
     }
   }
 
-  dst.ops->get_polling_fd (&dst, index, &fds[1].fd, &direction);
+  dst->ops->get_polling_fd (dst, index, &fds[1].fd, &direction);
   if (fds[1].fd >= 0) {
     switch (direction) {
     case LIBNBD_AIO_DIRECTION_READ:
@@ -311,24 +315,24 @@ poll_both_ends (uintptr_t index)
 
   if (fds[0].fd >= 0) {
     if ((fds[0].revents & (POLLIN | POLLHUP)) != 0)
-      src.ops->asynch_notify_read (&src, index);
+      src->ops->asynch_notify_read (src, index);
     else if ((fds[0].revents & POLLOUT) != 0)
-      src.ops->asynch_notify_write (&src, index);
+      src->ops->asynch_notify_write (src, index);
     else if ((fds[0].revents & (POLLERR | POLLNVAL)) != 0) {
       errno = ENOTCONN;
-      perror (src.name);
+      perror (src->name);
       exit (EXIT_FAILURE);
     }
   }
 
   if (fds[1].fd >= 0) {
     if ((fds[1].revents & (POLLIN | POLLHUP)) != 0)
-      dst.ops->asynch_notify_read (&dst, index);
+      dst->ops->asynch_notify_read (dst, index);
     else if ((fds[1].revents & POLLOUT) != 0)
-      dst.ops->asynch_notify_write (&dst, index);
+      dst->ops->asynch_notify_write (dst, index);
     else if ((fds[1].revents & (POLLERR | POLLNVAL)) != 0) {
       errno = ENOTCONN;
-      perror (dst.name);
+      perror (dst->name);
       exit (EXIT_FAILURE);
     }
   }
@@ -373,15 +377,21 @@ finished_read (void *vp, int *error)
 {
   struct command *command = vp;
 
+  if (*error) {
+    fprintf (stderr, "read at offset %" PRId64 " failed: %s\n",
+             command->offset, strerror (*error));
+    exit (EXIT_FAILURE);
+  }
+
   if (allocated || sparse_size == 0) {
     /* If sparseness detection (see below) is turned off then we write
      * the whole command.
      */
-    dst.ops->asynch_write (&dst, command,
-                           (nbd_completion_callback) {
-                             .callback = free_command,
-                             .user_data = command,
-                           });
+    dst->ops->asynch_write (dst, command,
+                            (nbd_completion_callback) {
+                              .callback = free_command,
+                              .user_data = command,
+                            });
   }
   else {                               /* Sparseness detection. */
     const uint64_t start = command->offset;
@@ -390,11 +400,12 @@ finished_read (void *vp, int *error)
     bool last_is_hole = false;
     uint64_t i;
     struct command *newcommand;
+    int dummy = 0;
 
     /* Iterate over whole blocks in the command, starting on a block
      * boundary.
      */
-    for (i = ROUND_UP (start, sparse_size);
+    for (i = MIN (ROUND_UP (start, sparse_size), end);
          i + sparse_size <= end;
          i += sparse_size) {
       if (is_zero (slice_ptr (command->slice) + i-start, sparse_size)) {
@@ -408,11 +419,11 @@ finished_read (void *vp, int *error)
             newcommand = copy_subcommand (command,
                                           last_offset, i - last_offset,
                                           false);
-            dst.ops->asynch_write (&dst, newcommand,
-                                   (nbd_completion_callback) {
-                                     .callback = free_command,
-                                     .user_data = newcommand,
-                                   });
+            dst->ops->asynch_write (dst, newcommand,
+                                    (nbd_completion_callback) {
+                                      .callback = free_command,
+                                      .user_data = newcommand,
+                                    });
           }
           /* Start the new hole. */
           last_offset = i;
@@ -445,11 +456,11 @@ finished_read (void *vp, int *error)
         newcommand = copy_subcommand (command,
                                       last_offset, i - last_offset,
                                       false);
-        dst.ops->asynch_write (&dst, newcommand,
-                               (nbd_completion_callback) {
-                                 .callback = free_command,
-                                 .user_data = newcommand,
-                               });
+        dst->ops->asynch_write (dst, newcommand,
+                                (nbd_completion_callback) {
+                                  .callback = free_command,
+                                  .user_data = newcommand,
+                                });
       }
       else {
         newcommand = copy_subcommand (command,
@@ -462,17 +473,17 @@ finished_read (void *vp, int *error)
     /* There may be an unaligned tail, so write that. */
     if (end - i > 0) {
       newcommand = copy_subcommand (command, i, end - i, false);
-      dst.ops->asynch_write (&dst, newcommand,
-                             (nbd_completion_callback) {
-                               .callback = free_command,
-                               .user_data = newcommand,
-                             });
+      dst->ops->asynch_write (dst, newcommand,
+                              (nbd_completion_callback) {
+                                .callback = free_command,
+                                .user_data = newcommand,
+                              });
     }
 
     /* Free the original command since it has been split into
      * subcommands and the original is no longer needed.
      */
-    free_command (command, &errno);
+    free_command (command, &dummy);
   }
 
   return 1; /* auto-retires the command */
@@ -485,11 +496,12 @@ finished_read (void *vp, int *error)
  * --destination-is-zero:
  *                 do nothing
  *
- * --allocated:    we must write zeroes either using an efficient
+ * --allocated:    write zeroes allocating space using an efficient
  *                 zeroing command or writing a command of zeroes
  *
- * (neither flag)  try trimming if supported, else write zeroes
- *                 as above
+ * (neither flag)  write zeroes punching a hole using an efficient
+ *                 zeroing command or fallback to writing a command
+ *                 of zeroes.
  *
  * This takes over ownership of the command and frees it eventually.
  */
@@ -497,32 +509,26 @@ static void
 fill_dst_range_with_zeroes (struct command *command)
 {
   char *data;
+  size_t data_size;
+  int dummy = 0;
 
   if (destination_is_zero)
     goto free_and_return;
 
-  if (!allocated) {
-    /* Try trimming. */
-    if (dst.ops->asynch_trim (&dst, command,
-                              (nbd_completion_callback) {
-                                .callback = free_command,
-                                .user_data = command,
-                              }))
-      return;
-  }
-
   /* Try efficient zeroing. */
-  if (dst.ops->asynch_zero (&dst, command,
-                            (nbd_completion_callback) {
-                              .callback = free_command,
-                              .user_data = command,
-                            }))
+  if (dst->ops->asynch_zero (dst, command,
+                             (nbd_completion_callback) {
+                               .callback = free_command,
+                               .user_data = command,
+                             },
+                             allocated))
     return;
 
   /* Fall back to loop writing zeroes.  This is going to be slow
    * anyway, so do it synchronously. XXX
    */
-  data = calloc (1, MAX_REQUEST_SIZE);
+  data_size = MIN (request_size, command->slice.len);
+  data = calloc (1, data_size);
   if (!data) {
     perror ("calloc");
     exit (EXIT_FAILURE);
@@ -530,17 +536,17 @@ fill_dst_range_with_zeroes (struct command *command)
   while (command->slice.len > 0) {
     size_t len = command->slice.len;
 
-    if (len > MAX_REQUEST_SIZE)
-      len = MAX_REQUEST_SIZE;
+    if (len > data_size)
+      len = data_size;
 
-    dst.ops->synch_write (&dst, data, len, command->offset);
+    dst->ops->synch_write (dst, data, len, command->offset);
     command->slice.len -= len;
     command->offset += len;
   }
   free (data);
 
  free_and_return:
-  free_command (command, &errno);
+  free_command (command, &dummy);
 }
 
 static int
@@ -549,6 +555,12 @@ free_command (void *vp, int *error)
   struct command *command = vp;
   struct buffer *buffer = command->slice.buffer;
 
+  if (*error) {
+    fprintf (stderr, "write at offset %" PRId64 " failed: %s\n",
+             command->offset, strerror (*error));
+    exit (EXIT_FAILURE);
+  }
+
   if (buffer != NULL) {
     if (--buffer->refs == 0) {
       free (buffer->data);
