@@ -28,23 +28,28 @@
 
 #include "nbdcopy.h"
 
-static char buf[MAX_REQUEST_SIZE];
-
 void
 synch_copying (void)
 {
   uint64_t offset = 0;
+  unsigned char *buf;
+
+  buf = malloc (request_size);
+  if (buf == NULL) {
+    perror ("malloc");
+    exit (EXIT_FAILURE);
+  }
 
   /* If the source size is unknown then we copy data and cannot use
    * extent information.
    */
-  if (src.size == -1) {
+  if (src->size == -1) {
     size_t r;
 
-    while ((r = src.ops->synch_read (&src, buf, sizeof buf, offset)) > 0) {
-      dst.ops->synch_write (&dst, buf, r, offset);
+    while ((r = src->ops->synch_read (src, buf, request_size, offset)) > 0) {
+      dst->ops->synch_write (dst, buf, r, offset);
       offset += r;
-      progress_bar (offset, src.size);
+      progress_bar (offset, src->size);
     }
   }
 
@@ -52,51 +57,53 @@ synch_copying (void)
    * blocks and use extent information to optimize the case.
    */
   else {
-    while (offset < src.size) {
+    while (offset < src->size) {
       extent_list exts = empty_vector;
-      uint64_t count = src.size - offset;
+      uint64_t count = src->size - offset;
       size_t i, r;
 
-      if (count > sizeof buf)
-        count = sizeof buf;
+      if (count > request_size)
+        count = request_size;
 
       if (extents)
-        src.ops->get_extents (&src, 0, offset, count, &exts);
+        src->ops->get_extents (src, 0, offset, count, &exts);
       else
-        default_get_extents (&src, 0, offset, count, &exts);
+        default_get_extents (src, 0, offset, count, &exts);
 
       for (i = 0; i < exts.size; ++i) {
         assert (exts.ptr[i].length <= count);
 
-        if (exts.ptr[i].hole) {
-          if (!dst.ops->synch_trim (&dst, offset, exts.ptr[i].length) &&
-              !dst.ops->synch_zero (&dst, offset, exts.ptr[i].length)) {
-            /* If neither trimming nor efficient zeroing are possible,
-             * write zeroes the hard way.
+        if (exts.ptr[i].zero) {
+          if (!dst->ops->synch_zero (dst, offset, exts.ptr[i].length, false) &&
+              !dst->ops->synch_zero (dst, offset, exts.ptr[i].length, true)) {
+            /* If efficient zeroing (punching a hole or allocating
+             * space) are possible, write zeroes the hard way.
              */
             memset (buf, 0, exts.ptr[i].length);
-            dst.ops->synch_write (&dst, buf, exts.ptr[i].length, offset);
+            dst->ops->synch_write (dst, buf, exts.ptr[i].length, offset);
           }
           offset += exts.ptr[i].length;
         }
         else /* data */ {
-          r = src.ops->synch_read (&src, buf, exts.ptr[i].length, offset);
+          r = src->ops->synch_read (src, buf, exts.ptr[i].length, offset);
 
           /* These cases should never happen unless the file is
            * truncated underneath us.
            */
           if (r == 0 || r < exts.ptr[i].length) {
-            fprintf (stderr, "%s: unexpected end of file\n", src.name);
+            fprintf (stderr, "%s: unexpected end of file\n", src->name);
             exit (EXIT_FAILURE);
           }
 
-          dst.ops->synch_write (&dst, buf, r, offset);
+          dst->ops->synch_write (dst, buf, r, offset);
           offset += r;
-          progress_bar (offset, src.size);
+          progress_bar (offset, src->size);
         }
       }
 
       free (exts.ptr);
     } /* while */
   }
+
+  free (buf);
 }
