@@ -1,5 +1,5 @@
 /* NBD client library in userspace
- * Copyright (C) 2013-2020 Red Hat Inc.
+ * Copyright (C) 2013-2021 Red Hat Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -28,58 +28,42 @@
 #include <string.h>
 #include <getopt.h>
 #include <limits.h>
-#include <fcntl.h>
 #include <unistd.h>
 #include <errno.h>
-#include <signal.h>
+#include <assert.h>
 #include <time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 
-#define FUSE_USE_VERSION 26
-
-#include <fuse.h>
-#ifdef HAVE_FUSE_LOWLEVEL_H
-#include <fuse_lowlevel.h>
-#endif
-
 #include <libnbd.h>
 
-#define MAX_REQUEST_SIZE (32 * 1024 * 1024)
+#include "version.h"
+#include "nbdfuse.h"
 
-static struct nbd_handle *nbd;
-static bool readonly;
-static char *mountpoint, *filename;
+handles nbd = empty_vector;
+unsigned connections = 4;
+bool readonly;
+bool file_mode = false;
+struct timespec start_t;
+char *filename;
+uint64_t size;
+bool verbose = false;
+
+static char *mountpoint;
 static const char *pidfile;
 static char *fuse_options;
-static struct fuse_chan *ch;
 static struct fuse *fuse;
-static struct timespec start_t;
-static uint64_t size;
+static struct fuse_session *fuse_session;
 
-static int nbdfuse_getattr (const char *path, struct stat *stbuf);
-static int nbdfuse_readdir (const char *path, void *buf,
-                            fuse_fill_dir_t filler,
-                            off_t offset, struct fuse_file_info *fi);
-static int nbdfuse_open (const char *path, struct fuse_file_info *fi);
-static int nbdfuse_read (const char *path, char *buf,
-                         size_t count, off_t offset,
-                         struct fuse_file_info *fi);
-static int nbdfuse_write (const char *path, const char *buf,
-                          size_t count, off_t offset,
-                          struct fuse_file_info *fi);
-static int nbdfuse_fsync (const char *path, int datasync,
-                          struct fuse_file_info *fi);
-static int nbdfuse_release (const char *path, struct fuse_file_info *fi);
-
-static struct fuse_operations fuse_operations = {
-  .getattr           = nbdfuse_getattr,
-  .readdir           = nbdfuse_readdir,
-  .open              = nbdfuse_open,
-  .read              = nbdfuse_read,
-  .write             = nbdfuse_write,
-  .fsync             = nbdfuse_fsync,
-  .release           = nbdfuse_release,
+enum mode {
+  MODE_URI,                  /* URI */
+  MODE_COMMAND,              /* --command */
+  MODE_FD,                   /* --fd */
+  MODE_SQUARE_BRACKET,       /* [ CMD ], same as --socket-activation*/
+  MODE_SOCKET_ACTIVATION,    /* --socket-activation */
+  MODE_TCP,                  /* --tcp */
+  MODE_UNIX,                 /* --unix */
+  MODE_VSOCK,                /* --vsock */
 };
 
 static void __attribute__((noreturn))
@@ -90,30 +74,36 @@ usage (FILE *fp, int exitcode)
 "Mount NBD server as a virtual file:\n"
 "\n"
 #ifdef HAVE_LIBXML2
-"    nbdfuse [-o FUSE-OPTION] [-P PIDFILE] [-r] MOUNTPOINT[/FILENAME] URI\n"
+"    nbdfuse [-C N|--connections N] [-d] [-o FUSE-OPTION] [-P PIDFILE]\n"
+"            [-r] [-s] [-v]\n"
+"            MOUNTPOINT[/FILENAME] URI\n"
 "\n"
 "Other modes:\n"
 "\n"
 #endif
+"    nbdfuse MOUNTPOINT[/FILENAME] [ CMD [ARGS ...] ]\n"
 "    nbdfuse MOUNTPOINT[/FILENAME] --command CMD [ARGS ...]\n"
-"    nbdfuse MOUNTPOINT[/FILENAME] --socket-activation CMD [ARGS ...]\n"
 "    nbdfuse MOUNTPOINT[/FILENAME] --fd N\n"
 "    nbdfuse MOUNTPOINT[/FILENAME] --tcp HOST PORT\n"
 "    nbdfuse MOUNTPOINT[/FILENAME] --unix SOCKET\n"
 "    nbdfuse MOUNTPOINT[/FILENAME] --vsock CID PORT\n"
 "\n"
+"To unmount:\n"
+"\n"
+"    fusermount3 -u MOUNTPOINT\n"
+"\n"
+"Other options:\n"
+"\n"
+"    nbdfuse --help\n"
+"    nbdfuse --fuse-help\n"
+"    nbdfuse -V|--version\n"
+"\n"
 "Please read the nbdfuse(1) manual page for full usage.\n"
 "\n"
 );
   exit (exitcode);
 }
 
-static void
-display_version (void)
-{
-  printf ("%s %s\n", PACKAGE_NAME, PACKAGE_VERSION);
-}
-
 static void
 fuse_help (const char *prog)
 {
@@ -133,44 +123,79 @@ is_directory (const char *path)
   return S_ISDIR (statbuf.st_mode);
 }
 
+static bool
+is_regular_file (const char *path)
+{
+  struct stat statbuf;
+
+  if (stat (path, &statbuf) == -1)
+    return false;
+  return S_ISREG (statbuf.st_mode);
+}
+
+/* Which modes support multi-conn?  We cannot connect multiple times
+ * to subprocesses (since we'd have to launch multiple subprocesses).
+ */
+static bool
+mode_is_multi_conn_compatible (enum mode mode)
+{
+  switch (mode) {
+  case MODE_COMMAND:
+  case MODE_SQUARE_BRACKET:
+  case MODE_SOCKET_ACTIVATION:
+  case MODE_FD:
+    return false;
+  case MODE_URI:
+  case MODE_TCP:
+  case MODE_UNIX:
+  case MODE_VSOCK:
+    return true;
+  default:
+    abort ();
+  }
+}
+
+static struct nbd_handle *create_and_connect (enum mode mode,
+                                              int argc, char **argv);
+
 int
 main (int argc, char *argv[])
 {
-  enum {
-    MODE_URI,
-    MODE_COMMAND,
-    MODE_FD,
-    MODE_SOCKET_ACTIVATION,
-    MODE_TCP,
-    MODE_UNIX,
-    MODE_VSOCK,
-  } mode = MODE_URI;
+  enum mode mode = MODE_URI;
   enum {
     HELP_OPTION = CHAR_MAX + 1,
     FUSE_HELP_OPTION,
+    LONG_OPTIONS,
+    SHORT_OPTIONS,
   };
   /* Note the "+" means we stop processing as soon as we get to the
    * first non-option argument (the mountpoint) and then we parse the
    * rest of the command line without getopt.
    */
-  const char *short_options = "+o:P:rV";
+  const char *short_options = "+C:do:P:rsvV";
   const struct option long_options[] = {
     { "fuse-help",          no_argument,       NULL, FUSE_HELP_OPTION },
     { "help",               no_argument,       NULL, HELP_OPTION },
+    { "long-options",       no_argument,       NULL, LONG_OPTIONS },
+    { "connections",        required_argument, NULL, 'C' },
     { "pidfile",            required_argument, NULL, 'P' },
     { "pid-file",           required_argument, NULL, 'P' },
     { "readonly",           no_argument,       NULL, 'r' },
     { "read-only",          no_argument,       NULL, 'r' },
+    { "short-options",      no_argument,       NULL, SHORT_OPTIONS },
+    { "verbose",            no_argument,       NULL, 'v' },
     { "version",            no_argument,       NULL, 'V' },
 
     { NULL }
   };
-  int c, fd, r;
-  uint32_t cid, port;
+  int c, r;
+  size_t i;
+  bool singlethread = false;
+  struct nbd_handle *h;
   int64_t ssize;
   const char *s;
   struct fuse_args fuse_args = FUSE_ARGS_INIT (0, NULL);
-  struct sigaction sa;
+  struct fuse_loop_config fuse_loop_config;
   FILE *fp;
 
   for (;;) {
@@ -186,6 +211,30 @@ main (int argc, char *argv[])
       fuse_help (argv[0]);
       exit (EXIT_SUCCESS);
 
+    case LONG_OPTIONS:
+      for (i = 0; long_options[i].name != NULL; ++i) {
+        if (strcmp (long_options[i].name, "long-options") != 0 &&
+            strcmp (long_options[i].name, "short-options") != 0)
+          printf ("--%s\n", long_options[i].name);
+      }
+      exit (EXIT_SUCCESS);
+
+    case SHORT_OPTIONS:
+      for (i = 0; short_options[i]; ++i) {
+        if (short_options[i] != ':' && short_options[i] != '+')
+          printf ("-%c\n", short_options[i]);
+      }
+      exit (EXIT_SUCCESS);
+
+    case 'C':
+      if (sscanf (optarg, "%u", &connections) != 1 ||
+          connections < 1 || connections > 1024) {
+        fprintf (stderr, "%s: --connections parameter must be an unsigned integer >= 1\n",
+                 argv[0]);
+        exit (EXIT_FAILURE);
+      }
+      break;
+
     case 'o':
       fuse_opt_add_opt_escaped (&fuse_options, optarg);
       break;
@@ -198,8 +247,16 @@ main (int argc, char *argv[])
       readonly = true;
       break;
 
+    case 's':
+      singlethread = true;
+      break;
+
+    case 'v': case 'd':
+      verbose = true;
+      break;
+
     case 'V':
-      display_version ();
+      display_version ("nbdfuse");
       exit (EXIT_SUCCESS);
 
     default:
@@ -213,21 +270,36 @@ main (int argc, char *argv[])
   if (argc - optind < 2)
     usage (stderr, EXIT_FAILURE);
 
-  /* Parse and check the mountpoint.  It might be MOUNTPOINT or
-   * MOUNTPOINT/FILENAME.  In either case MOUNTPOINT must be an
-   * existing directory.
+  /* Parse and check the mountpoint.  It might be MOUNTPOINT (file),
+   * MOUNTPOINT (directory), or MOUNTPOINT/FILENAME.
    */
   s = argv[optind++];
-  if (is_directory (s)) {
+  if (is_regular_file (s)) {    /* MOUNTPOINT (file) */
+    const char *p;
+
+    file_mode = true;
     mountpoint = strdup (s);
-    filename = strdup ("nbd");
-    if (mountpoint == NULL || filename == NULL) {
+    if (mountpoint == NULL) {
     strdup_error:
       perror ("strdup");
       exit (EXIT_FAILURE);
     }
+    p = strrchr (s, '/');
+    if (!p)
+      filename = strdup (s);
+    else {
+      if (strlen (p+1) == 0) goto mp_error; /* probably can't happen */
+      filename = strdup (p+1);
+    }
+    if (!filename) goto strdup_error;
   }
-  else {
+  else if (is_directory (s)) {  /* MOUNTPOINT/nbd */
+    mountpoint = strdup (s);
+    filename = strdup ("nbd");
+    if (mountpoint == NULL || filename == NULL)
+      goto strdup_error;
+  }
+  else {                        /* MOUNTPOINT/FILENAME */
     const char *p = strrchr (s, '/');
 
     if (p == NULL) {
@@ -251,6 +323,10 @@ main (int argc, char *argv[])
     mode = MODE_COMMAND;
     optind++;
   }
+  else if (strcmp (argv[optind], "[") == 0) {
+    mode = MODE_SQUARE_BRACKET;
+    optind++;
+  }
   else if (strcmp (argv[optind], "--socket-activation") == 0 ||
            strcmp (argv[optind], "--systemd-socket-activation") == 0) {
     mode = MODE_SOCKET_ACTIVATION;
@@ -308,92 +384,59 @@ main (int argc, char *argv[])
     if (argc - optind < 1)
       usage (stderr, EXIT_FAILURE);
     break;
+  case MODE_SQUARE_BRACKET:
+    if (argc - optind < 2 || strcmp (argv[argc-1], "]") != 0)
+      usage (stderr, EXIT_FAILURE);
+    break;
   }
   /* At this point we know the command line is valid, and so can start
    * opening FUSE and libnbd.
    */
 
-  /* Create the libnbd handle. */
-  nbd = nbd_create ();
-  if (nbd == NULL) {
-    fprintf (stderr, "%s\n", nbd_get_error ());
+  /* Create the libnbd handle and connect to it. */
+  h = create_and_connect (mode, argc, argv);
+  if (handles_append (&nbd, h) == -1) {
+    perror ("realloc");
     exit (EXIT_FAILURE);
   }
 
-  /* Connect to the NBD server synchronously. */
-  switch (mode) {
-  case MODE_URI:
-    if (nbd_connect_uri (nbd, argv[optind]) == -1) {
-      fprintf (stderr, "%s\n", nbd_get_error ());
+  /* If the server supports multi-conn, and we are able to, try to
+   * open more handles.
+   */
+  if (connections > 1 &&
+      mode_is_multi_conn_compatible (mode) &&
+      nbd_can_multi_conn (nbd.ptr[0]) >= 1) {
+    if (handles_reserve (&nbd, connections-1) == -1) {
+      perror ("realloc");
       exit (EXIT_FAILURE);
     }
-    break;
-
-  case MODE_COMMAND:
-    if (nbd_connect_command (nbd, &argv[optind]) == -1) {
-      fprintf (stderr, "%s\n", nbd_get_error ());
-      exit (EXIT_FAILURE);
+    for (i = 2; i <= connections; ++i) {
+      h = create_and_connect (mode, argc, argv);
+      handles_append (&nbd, h); /* reserved above, so can't fail */
     }
-    break;
-
-  case MODE_SOCKET_ACTIVATION:
-    if (nbd_connect_systemd_socket_activation (nbd, &argv[optind]) == -1) {
-      fprintf (stderr, "%s\n", nbd_get_error ());
-      exit (EXIT_FAILURE);
-    }
-    break;
-
-  case MODE_FD:
-    if (sscanf (argv[optind], "%d", &fd) != 1) {
-      fprintf (stderr, "%s: could not parse file descriptor: %s\n\n",
-               argv[0], argv[optind]);
-      exit (EXIT_FAILURE);
-    }
-    if (nbd_connect_socket (nbd, fd) == -1) {
-      fprintf (stderr, "%s\n", nbd_get_error ());
-      exit (EXIT_FAILURE);
-    }
-    break;
-
-  case MODE_TCP:
-    if (nbd_connect_tcp (nbd, argv[optind], argv[optind+1]) == -1) {
-      fprintf (stderr, "%s\n", nbd_get_error ());
-      exit (EXIT_FAILURE);
-    }
-    break;
-
-  case MODE_UNIX:
-    if (nbd_connect_unix (nbd, argv[optind]) == -1) {
-      fprintf (stderr, "%s\n", nbd_get_error ());
-      exit (EXIT_FAILURE);
-    }
-    break;
-
-  case MODE_VSOCK:
-    if (sscanf (argv[optind], "%" SCNu32, &cid) != 1) {
-      fprintf (stderr, "%s: could not parse vsock cid: %s\n\n",
-               argv[0], argv[optind]);
-      exit (EXIT_FAILURE);
-    }
-    if (sscanf (argv[optind+1], "%" SCNu32, &port) != 1) {
-      fprintf (stderr, "%s: could not parse vsock port: %s\n\n",
-               argv[0], argv[optind]);
-      exit (EXIT_FAILURE);
-    }
-    if (nbd_connect_vsock (nbd, cid, port) == -1) {
-      fprintf (stderr, "%s\n", nbd_get_error ());
-      exit (EXIT_FAILURE);
-    }
-    break;
   }
+  connections = (unsigned) nbd.size;
+  if (verbose)
+    fprintf (stderr, "nbdfuse: connections=%u\n", connections);
 
-  ssize = nbd_get_size (nbd);
+  ssize = nbd_get_size (nbd.ptr[0]);
   if (ssize == -1) {
     fprintf (stderr, "%s\n", nbd_get_error ());
     exit (EXIT_FAILURE);
   }
   size = (uint64_t) ssize;
 
+  /* If the remote NBD server is readonly, then act as if the '-r'
+   * flag was given on the nbdfuse command line.
+   */
+  if (nbd_is_read_only (nbd.ptr[0]) > 0)
+    readonly = true;
+
+  /* Create the background threads which are used to dispatch NBD
+   * operations.
+   */
+  start_operations_threads ();
+
   /* This is just used to give an unchanging time when they stat in
    * the mountpoint.
    */
@@ -413,43 +456,26 @@ main (int argc, char *argv[])
   }
 
   /* Create the FUSE mountpoint. */
-  ch = fuse_mount (mountpoint, &fuse_args);
-  if (ch == NULL) {
-    fprintf (stderr,
-             "%s: fuse_mount failed: see error messages above", argv[0]);
-    exit (EXIT_FAILURE);
-  }
-
-  /* Set FD_CLOEXEC on the channel.  Some versions of libfuse don't do
-   * this.
-   */
-  fd = fuse_chan_fd (ch);
-  if (fd >= 0) {
-    int flags = fcntl (fd, F_GETFD, 0);
-    if (flags >= 0)
-      fcntl (fd, F_SETFD, flags | FD_CLOEXEC);
-  }
-
-  /* Create the FUSE handle. */
-  fuse = fuse_new (ch, &fuse_args,
-                   &fuse_operations, sizeof fuse_operations, NULL);
+  fuse = fuse_new (&fuse_args,
+                   &nbdfuse_operations, sizeof nbdfuse_operations, NULL);
   if (!fuse) {
     perror ("fuse_new");
     exit (EXIT_FAILURE);
   }
   fuse_opt_free_args (&fuse_args);
 
-  /* Catch signals since they can leave the mountpoint in a funny
-   * state.  To exit the program callers must use ‘fusermount -u’.  We
-   * also must be careful not to call exit(2) in this program until we
-   * have unmounted the filesystem below.
+  fuse_session = fuse_get_session (fuse);
+
+  fuse_set_signal_handlers (fuse_session);
+
+  /* After successful fuse_mount we must be careful not to call
+   * exit(2) in this program until we have unmounted the filesystem
+   * below.
    */
-  memset (&sa, 0, sizeof sa);
-  sa.sa_handler = SIG_IGN;
-  sa.sa_flags = SA_RESTART;
-  sigaction (SIGPIPE, &sa, NULL);
-  sigaction (SIGINT, &sa, NULL);
-  sigaction (SIGQUIT, &sa, NULL);
+  if (fuse_mount (fuse, mountpoint) == -1) {
+    perror ("fuse_mount");
+    exit (EXIT_FAILURE);
+  }
 
   /* Ready to serve, write pidfile. */
   if (pidfile) {
@@ -461,16 +487,32 @@ main (int argc, char *argv[])
   }
 
   /* Enter the main loop. */
-  r = fuse_loop (fuse);
-  if (r != 0)
+  if (!singlethread) {
+    memset (&fuse_loop_config, 0, sizeof fuse_loop_config);
+    fuse_loop_config.clone_fd = 0;
+    fuse_loop_config.max_idle_threads = 10;
+    r = fuse_loop_mt (fuse, &fuse_loop_config);
+  }
+  else {
+    r = fuse_loop (fuse);
+  }
+  if (r < 0) {
+    errno = -r;
     perror ("fuse_loop");
+  }
+  else if (r > 0) {
+    fprintf (stderr, "%s: fuse_loop: fuse loop terminated by signal %d\n",
+             argv[0], r);
+  }
 
   /* Close FUSE. */
-  fuse_unmount (mountpoint, ch);
+  fuse_remove_signal_handlers (fuse_session);
+  fuse_unmount (fuse);
   fuse_destroy (fuse);
 
-  /* Close NBD handle. */
-  nbd_close (nbd);
+  /* Close NBD handle(s). */
+  for (i = 0; i < nbd.size; ++i)
+    nbd_close (nbd.ptr[i]);
 
   free (mountpoint);
   free (filename);
@@ -479,161 +521,97 @@ main (int argc, char *argv[])
   exit (r == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
 
-/* Wraps calls to libnbd functions and automatically checks for error,
- * returning errors in the format required by FUSE.  It also prints
- * out the full error message on stderr, so that we don't lose it.
+/* Called from main() above to create an NBD handle and connect to it.
+ * For multi-conn, this may be called several times.
  */
-#define CHECK_NBD_ERROR(CALL)                                   \
-  do { if ((CALL) == -1) return check_nbd_error (); } while (0)
-static int
-check_nbd_error (void)
+static struct nbd_handle *
+create_and_connect (enum mode mode, int argc, char **argv)
 {
-  int err;
+  int fd;
+  uint32_t cid, port;
+  struct nbd_handle *h;
 
-  fprintf (stderr, "%s\n", nbd_get_error ());
-  err = nbd_get_errno ();
-  if (err != 0)
-    return -err;
-  else
-    return -EIO;
-}
-
-static int
-nbdfuse_getattr (const char *path, struct stat *statbuf)
-{
-  const int mode = readonly ? 0444 : 0666;
-
-  memset (statbuf, 0, sizeof (struct stat));
-
-  /* We're probably making some Linux-specific assumptions here, but
-   * this file is not usually compiled on non-Linux systems (perhaps
-   * on OpenBSD?).  XXX
-   */
-  statbuf->st_atim = start_t;
-  statbuf->st_mtim = start_t;
-  statbuf->st_ctim = start_t;
-  statbuf->st_uid = geteuid ();
-  statbuf->st_gid = getegid ();
-
-  if (strcmp (path, "/") == 0) {
-    /* getattr "/" */
-    statbuf->st_mode = S_IFDIR | (mode & 0111);
-    statbuf->st_nlink = 2;
+  /* Create the libnbd handle. */
+  h = nbd_create ();
+  if (h == NULL) {
+    fprintf (stderr, "%s\n", nbd_get_error ());
+    exit (EXIT_FAILURE);
   }
-  else if (path[0] == '/' && strcmp (path+1, filename) == 0) {
-    /* getattr "/filename" */
-    statbuf->st_mode = S_IFREG | mode;
-    statbuf->st_nlink = 1;
-    statbuf->st_size = size;
+  nbd_set_debug (h, verbose);
+
+  /* Connect to the NBD server synchronously. */
+  switch (mode) {
+  case MODE_URI:
+    if (nbd_connect_uri (h, argv[optind]) == -1) {
+      fprintf (stderr, "%s\n", nbd_get_error ());
+      exit (EXIT_FAILURE);
+    }
+    break;
+
+  case MODE_COMMAND:
+    if (nbd_connect_command (h, &argv[optind]) == -1) {
+      fprintf (stderr, "%s\n", nbd_get_error ());
+      exit (EXIT_FAILURE);
+    }
+    break;
+
+  case MODE_SQUARE_BRACKET:
+    /* This is the same as MODE_SOCKET_ACTIVATION but we must eat the
+     * closing square bracket on the command line.
+     */
+    assert (strcmp (argv[argc-1], "]") == 0); /* checked above */
+    argv[argc-1] = NULL;
+    /*FALLTHROUGH*/
+  case MODE_SOCKET_ACTIVATION:
+    if (nbd_connect_systemd_socket_activation (h, &argv[optind]) == -1) {
+      fprintf (stderr, "%s\n", nbd_get_error ());
+      exit (EXIT_FAILURE);
+    }
+    break;
+
+  case MODE_FD:
+    if (sscanf (argv[optind], "%d", &fd) != 1) {
+      fprintf (stderr, "%s: could not parse file descriptor: %s\n\n",
+               argv[0], argv[optind]);
+      exit (EXIT_FAILURE);
+    }
+    if (nbd_connect_socket (h, fd) == -1) {
+      fprintf (stderr, "%s\n", nbd_get_error ());
+      exit (EXIT_FAILURE);
+    }
+    break;
+
+  case MODE_TCP:
+    if (nbd_connect_tcp (h, argv[optind], argv[optind+1]) == -1) {
+      fprintf (stderr, "%s\n", nbd_get_error ());
+      exit (EXIT_FAILURE);
+    }
+    break;
+
+  case MODE_UNIX:
+    if (nbd_connect_unix (h, argv[optind]) == -1) {
+      fprintf (stderr, "%s\n", nbd_get_error ());
+      exit (EXIT_FAILURE);
+    }
+    break;
+
+  case MODE_VSOCK:
+    if (sscanf (argv[optind], "%" SCNu32, &cid) != 1) {
+      fprintf (stderr, "%s: could not parse vsock cid: %s\n\n",
+               argv[0], argv[optind]);
+      exit (EXIT_FAILURE);
+    }
+    if (sscanf (argv[optind+1], "%" SCNu32, &port) != 1) {
+      fprintf (stderr, "%s: could not parse vsock port: %s\n\n",
+               argv[0], argv[optind]);
+      exit (EXIT_FAILURE);
+    }
+    if (nbd_connect_vsock (h, cid, port) == -1) {
+      fprintf (stderr, "%s\n", nbd_get_error ());
+      exit (EXIT_FAILURE);
+    }
+    break;
   }
-  else
-    return -ENOENT;
 
-  return 0;
-}
-
-static int
-nbdfuse_readdir (const char *path, void *buf,
-                 fuse_fill_dir_t filler,
-                 off_t offset, struct fuse_file_info *fi)
-{
-  if (strcmp (path, "/") != 0)
-    return -ENOENT;
-
-  filler (buf, ".", NULL, 0);
-  filler (buf, "..", NULL, 0);
-  filler (buf, filename, NULL, 0);
-
-  return 0;
-}
-
-/* This function checks the O_RDONLY/O_RDWR flags passed to the
- * open(2) call, so we have to check the open mode is compatible with
- * the readonly flag.
- */
-static int
-nbdfuse_open (const char *path, struct fuse_file_info *fi)
-{
-  if (path[0] != '/' || strcmp (path+1, filename) != 0)
-    return -ENOENT;
-
-  if (readonly && (fi->flags & O_ACCMODE) != O_RDONLY)
-    return -EACCES;
-
-  return 0;
-}
-
-static int
-nbdfuse_read (const char *path, char *buf,
-              size_t count, off_t offset,
-              struct fuse_file_info *fi)
-{
-  if (path[0] != '/' || strcmp (path+1, filename) != 0)
-    return -ENOENT;
-
-  if (offset >= size)
-    return 0;
-
-  if (count > MAX_REQUEST_SIZE)
-    count = MAX_REQUEST_SIZE;
-
-  if (offset + count > size)
-    count = size - offset;
-
-  CHECK_NBD_ERROR (nbd_pread (nbd, buf, count, offset, 0));
-
-  return (int) count;
-}
-
-static int
-nbdfuse_write (const char *path, const char *buf,
-               size_t count, off_t offset,
-               struct fuse_file_info *fi)
-{
-  /* Probably shouldn't happen because of nbdfuse_open check. */
-  if (readonly)
-    return -EACCES;
-
-  if (path[0] != '/' || strcmp (path+1, filename) != 0)
-    return -ENOENT;
-
-  if (offset >= size)
-    return 0;
-
-  if (count > MAX_REQUEST_SIZE)
-    count = MAX_REQUEST_SIZE;
-
-  if (offset + count > size)
-    count = size - offset;
-
-  CHECK_NBD_ERROR (nbd_pwrite (nbd, buf, count, offset, 0));
-
-  return (int) count;
-}
-
-static int
-nbdfuse_fsync (const char *path, int datasync, struct fuse_file_info *fi)
-{
-  if (readonly)
-    return 0;
-
-  /* If the server doesn't support flush then the operation is
-   * silently ignored.
-   */
-  if (nbd_can_flush (nbd))
-    CHECK_NBD_ERROR (nbd_flush (nbd, 0));
-
-  return 0;
-}
-
-/* This is called on the last close of a file.  We do a flush here to
- * be on the safe side, but it's not strictly necessary.
- */
-static int
-nbdfuse_release (const char *path, struct fuse_file_info *fi)
-{
-  if (readonly)
-    return 0;
-
-  return nbdfuse_fsync (path, 0, fi);
+  return h;
 }
