@@ -61,6 +61,7 @@ let go_name_of_arg = function
   | UInt n -> n
   | UInt32 n -> n
   | UInt64 n -> n
+  | UIntPtr n -> n
 
 let go_arg_type = function
   | Bool _ -> "bool"
@@ -82,6 +83,7 @@ let go_arg_type = function
   | UInt _ -> "uint"
   | UInt32 _ -> "uint32"
   | UInt64 _ -> "uint64"
+  | UIntPtr _ -> "uint"
 
 let go_name_of_optarg = function
   | OClosure { cbname } -> sprintf "%sCallback" (camel_case cbname)
@@ -98,10 +100,11 @@ let go_ret_type = function
   | RCookie -> Some "uint64"
   | RSizeT -> Some "uint"
   | RString -> Some "*string"
-  (* RUInt returns (type, error) for consistency, but the error is
-   * always nil unless h is closed
+  (* RUInt | RUIntPtr returns (type, error) for consistency, but the
+   * error is always nil unless h is closed
    *)
   | RUInt -> Some "uint"
+  | RUIntPtr -> Some "uint"
   | REnum { enum_prefix } -> Some (camel_case enum_prefix)
   | RFlags { flag_prefix } -> Some (camel_case flag_prefix)
 
@@ -109,7 +112,7 @@ let go_ret_error = function
   | RErr -> None
   | RBool -> Some "false"
   | RStaticString | RString -> Some "nil"
-  | RFd | RInt | RInt64 | RCookie | RSizeT | RUInt
+  | RFd | RInt | RInt64 | RCookie | RSizeT | RUInt | RUIntPtr
   | REnum _ | RFlags _ -> Some "0"
 
 let go_ret_c_errcode = function
@@ -117,7 +120,7 @@ let go_ret_c_errcode = function
   | RStaticString -> Some "nil"
   | RErr | RFd | RInt | RInt64 | RCookie | RSizeT -> Some "-1"
   | RString -> Some "nil"
-  | RUInt | REnum _ | RFlags _ -> None
+  | RUInt | RUIntPtr | REnum _ | RFlags _ -> None
 
 (* We need a wrapper around every function (except Close) to
  * handle errors because cgo calls are sequence points and
@@ -172,8 +175,10 @@ let print_callback_wrapper { cbname; cbargs } =
   pr "void\n";
   pr "_nbd_%s_callback_free (void *user_data)\n" cbname;
   pr "{\n";
+  pr "  long *p = user_data;\n";
   pr "  extern void freeCallbackId (long);\n";
-  pr "  freeCallbackId ((long)user_data);\n";
+  pr "  freeCallbackId (*p);\n";
+  pr "  free (p);\n";
   pr "}\n";
   pr "\n"
 
@@ -252,8 +257,8 @@ let print_binding (name, { args; optargs; ret; shortdesc }) =
          cbname cbname;
        pr "    c_%s.free = (*[0]byte)(C._nbd_%s_callback_free)\n"
          cbname cbname;
-       pr "    c_%s.user_data = unsafe.Pointer (C.long_to_vp (C.long (registerCallbackId (%s))))\n"
-         cbname cbname
+       pr "    %s_cbid := registerCallbackId(%s)\n" cbname cbname;
+       pr "    c_%s.user_data = C.alloc_cbid(C.long(%s_cbid))\n" cbname cbname
     | Enum (n, _) ->
        pr "    c_%s := C.int (%s)\n" n n
     | Fd n ->
@@ -285,6 +290,8 @@ let print_binding (name, { args; optargs; ret; shortdesc }) =
        pr "    c_%s := C.uint32_t (%s)\n" n n
     | UInt64 n ->
        pr "    c_%s := C.uint64_t (%s)\n" n n
+    | UIntPtr n ->
+       pr "    c_%s := C.uintptr_t (%s)\n" n n
   ) args;
   if optargs <> [] then (
     List.iter (
@@ -303,8 +310,10 @@ let print_binding (name, { args; optargs; ret; shortdesc }) =
                cbname cbname;
              pr "            c_%s.free = (*[0]byte)(C._nbd_%s_callback_free)\n"
                cbname cbname;
-             pr "            c_%s.user_data = unsafe.Pointer (C.long_to_vp (C.long (registerCallbackId (optargs.%s))))\n"
-               cbname (go_name_of_optarg optarg)
+             pr "            %s_cbid := registerCallbackId(optargs.%s)\n"
+               cbname (go_name_of_optarg optarg);
+             pr "            c_%s.user_data = C.alloc_cbid(C.long(%s_cbid))\n"
+               cbname cbname
           | OFlags (n, _, _) ->
              pr "            c_%s = C.uint32_t (optargs.%s)\n"
                n (go_name_of_optarg optarg);
@@ -333,7 +342,7 @@ let print_binding (name, { args; optargs; ret; shortdesc }) =
     | SockAddrAndLen (n, len) -> pr ", c_%s, c_%s" n len
     | String n -> pr ", c_%s" n
     | StringList n -> pr ", &c_%s[0]" n
-    | UInt n -> pr ", c_%s" n
+    | UInt n | UIntPtr n -> pr ", c_%s" n
     | UInt32 n -> pr ", c_%s" n
     | UInt64 n -> pr ", c_%s" n
   ) args;
@@ -389,6 +398,8 @@ let print_binding (name, { args; optargs; ret; shortdesc }) =
       pr "    return &r, nil\n"
    | RUInt ->
       pr "    return uint (ret), nil\n"
+   | RUIntPtr ->
+      pr "    return uint (ret), nil\n"
    | REnum { enum_prefix } ->
       pr "    return %s (ret), nil\n" (camel_case enum_prefix)
    | RFlags { flag_prefix } ->
@@ -542,7 +553,7 @@ func copy_uint32_array (entries *C.uint32_t, count C.size_t) []uint32 {
       pr ") int\n";
       pr "\n";
       pr "//export %s_callback\n" cbname;
-      pr "func %s_callback (callbackid C.long" cbname;
+      pr "func %s_callback (callbackid *C.long" cbname;
       List.iter (
         fun cbarg ->
           pr ", ";
@@ -566,7 +577,7 @@ func copy_uint32_array (entries *C.uint32_t, count C.size_t) []uint32 {
           | CBArrayAndLen _ | CBMutable _ -> assert false
       ) cbargs;
       pr ") C.int {\n";
-      pr "    callbackFunc := getCallbackId (int (callbackid));\n";
+      pr "    callbackFunc := getCallbackId (int (*callbackid));\n";
       pr "    callback, ok := callbackFunc.(%sCallback);\n" uname;
       pr "    if !ok {\n";
       pr "        panic (\"inappropriate callback type\");\n";
@@ -662,15 +673,23 @@ let generate_golang_wrappers_h () =
 
 #include <stdio.h>
 #include <string.h>
+#include <assert.h>
 #include <errno.h>
 
 #include \"libnbd.h\"
 
-/* When calling callbacks we pass the callback ID (an int) in
- * the void *user_data field.  I couldn't work out how to do
- * this in golang, so this helper function does the cast.
+/* When calling callbacks we pass the callback ID (a golang int /
+ * C.long) in the void *user_data field.  We need to create a block
+ * to store the callback number.  This must be freed by C.free(vp)
  */
-static inline void *long_to_vp (long i) { return (void *)(intptr_t)i; }
+static inline void *
+alloc_cbid (long i)
+{
+  long *p = malloc (sizeof (long));
+  assert (p != NULL);
+  *p = i;
+  return p;
+}
 
 /* save_error is called from the same thread to make a copy
  * of the error which can later be retrieve from golang code
