@@ -1,6 +1,6 @@
 (* hey emacs, this is OCaml code: -*- tuareg -*- *)
 (* nbd client library in userspace: Python bindings
- * Copyright (C) 2013-2020 Red Hat Inc.
+ * Copyright (C) 2013-2021 Red Hat Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -41,6 +41,8 @@ struct py_aio_buffer {
 
 extern char **nbd_internal_py_get_string_list (PyObject *);
 extern void nbd_internal_py_free_string_list (char **);
+extern int nbd_internal_py_get_sockaddr (PyObject *,
+    struct sockaddr_storage *, socklen_t *);
 extern struct py_aio_buffer *nbd_internal_py_get_aio_buffer (PyObject *);
 
 static inline struct nbd_handle *
@@ -59,8 +61,10 @@ raise_exception ()
 {
   PyObject *args = Py_BuildValue (\"si\", nbd_get_error (), nbd_get_errno ());
 
-  if (args != NULL)
+  if (args != NULL) {
     PyErr_SetObject (nbd_internal_py_Error, args);
+    Py_DECREF (args);
+  }
 }
 
 ";
@@ -159,28 +163,42 @@ let print_python_closure_wrapper { cbname; cbargs } =
   pr "\n";
   pr "{\n";
   pr "  const struct user_data *data = user_data;\n";
-  pr "  int ret = 0;\n";
+  pr "  int ret = -1;\n";
   pr "\n";
   pr "  PyGILState_STATE py_save = PyGILState_UNLOCKED;\n";
   pr "  PyObject *py_args, *py_ret;\n";
+  List.iter (
+    function
+    | CBArrayAndLen (UInt32 n, _) ->
+       pr "  PyObject *py_%s = NULL;\n" n
+    | CBMutable (Int n) ->
+       pr "  PyObject *py_%s = NULL;\n" n
+    | _ -> ()
+  ) cbargs;
+  pr "\n";
   List.iter (
     function
     | CBArrayAndLen (UInt32 n, len) ->
-       pr "  PyObject *py_%s = PyList_New (%s);\n" n len;
-       pr "  for (size_t i = 0; i < %s; ++i)\n" len;
-       pr "    PyList_SET_ITEM (py_%s, i, PyLong_FromUnsignedLong (%s[i]));\n" n n
+       pr "  py_%s = PyList_New (%s);\n" n len;
+       pr "  if (!py_%s) { PyErr_PrintEx (0); goto out; }\n" n;
+       pr "  size_t i_%s;\n" n;
+       pr "  for (i_%s = 0; i_%s < %s; ++i_%s) {\n" n n len n;
+       pr "    PyObject *py_e_%s = PyLong_FromUnsignedLong (%s[i_%s]);\n" n n n;
+       pr "    if (!py_e_%s) { PyErr_PrintEx (0); goto out; }\n" n;
+       pr "    PyList_SET_ITEM (py_%s, i_%s, py_e_%s);\n" n n n;
+       pr "  }\n"
     | CBBytesIn _
     | CBInt _
     | CBInt64 _ -> ()
     | CBMutable (Int n) ->
        pr "  PyObject *py_%s_modname = PyUnicode_FromString (\"ctypes\");\n" n;
-       pr "  if (!py_%s_modname) { PyErr_PrintEx (0); return -1; }\n" n;
+       pr "  if (!py_%s_modname) { PyErr_PrintEx (0); goto out; }\n" n;
        pr "  PyObject *py_%s_mod = PyImport_Import (py_%s_modname);\n" n n;
        pr "  Py_DECREF (py_%s_modname);\n" n;
-       pr "  if (!py_%s_mod) { PyErr_PrintEx (0); return -1; }\n" n;
-       pr "  PyObject *py_%s = PyObject_CallMethod (py_%s_mod, \"c_int\", \"i\", *%s);\n" n n n;
+       pr "  if (!py_%s_mod) { PyErr_PrintEx (0); goto out; }\n" n;
+       pr "  py_%s = PyObject_CallMethod (py_%s_mod, \"c_int\", \"i\", *%s);\n" n n n;
        pr "  Py_DECREF (py_%s_mod);\n" n;
-       pr "  if (!py_%s) { PyErr_PrintEx (0); return -1; }\n" n;
+       pr "  if (!py_%s) { PyErr_PrintEx (0); goto out; }\n" n;
     | CBString _
     | CBUInt _
     | CBUInt64 _ -> ()
@@ -213,7 +231,7 @@ let print_python_closure_wrapper { cbname; cbargs } =
     | CBArrayAndLen _ | CBMutable _ -> assert false
   ) cbargs;
   pr ");\n";
-  pr "  Py_INCREF (py_args);\n";
+  pr "  if (!py_args) { PyErr_PrintEx (0); goto out; }\n";
   pr "\n";
   pr "  py_save = PyGILState_Ensure ();\n";
   pr "  py_ret = PyObject_CallObject (data->fn, py_args);\n";
@@ -235,19 +253,21 @@ let print_python_closure_wrapper { cbname; cbargs } =
   pr "      PyErr_Print ();\n";
   pr "      abort ();\n";
   pr "    }\n";
-  pr "    ret = -1;\n";
   pr "    PyErr_PrintEx (0); /* print exception */\n";
   pr "  };\n";
   pr "\n";
+  pr " out:\n";
   List.iter (
     function
     | CBArrayAndLen (UInt32 n, _) ->
-       pr "  Py_DECREF (py_%s);\n" n
+       pr "  Py_XDECREF (py_%s);\n" n
     | CBMutable (Int n) ->
-       pr "  PyObject *py_%s_ret = PyObject_GetAttrString (py_%s, \"value\");\n" n n;
-       pr "  *%s = PyLong_AsLong (py_%s_ret);\n" n n;
-       pr "  Py_DECREF (py_%s_ret);\n" n;
-       pr "  Py_DECREF (py_%s);\n" n
+       pr "  if (py_%s) {\n" n;
+       pr "    PyObject *py_%s_ret = PyObject_GetAttrString (py_%s, \"value\");\n" n n;
+       pr "    *%s = PyLong_AsLong (py_%s_ret);\n" n n;
+       pr "    Py_DECREF (py_%s_ret);\n" n;
+       pr "    Py_DECREF (py_%s);\n" n;
+       pr "  }\n"
     | CBBytesIn _
     | CBInt _ | CBInt64 _
     | CBString _
@@ -271,7 +291,7 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     function
     | Bool n -> pr "  int %s;\n" n
     | BytesIn (n, _) ->
-       pr "  Py_buffer %s;\n" n
+       pr "  Py_buffer %s = { .obj = NULL };\n" n
     | BytesOut (n, count) ->
        pr "  char *%s = NULL;\n" n;
        pr "  Py_ssize_t %s;\n" count
@@ -300,15 +320,14 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | SizeT n ->
        pr "  Py_ssize_t %s;\n" n
     | SockAddrAndLen (n, _) ->
-       pr "  /* XXX Complicated - Python uses a tuple of different\n";
-       pr "   * lengths for the different socket types.\n";
-       pr "   */\n";
-       pr "  PyObject *%s;\n" n
+       pr "  PyObject *%s;\n" n;
+       pr "  struct sockaddr_storage %s_sa;\n" n;
+       pr "  socklen_t %s_len;\n" n;
     | String n -> pr "  const char *%s;\n" n
     | StringList n ->
        pr "  PyObject *py_%s;\n" n;
        pr "  char **%s = NULL;\n" n
-    | UInt n -> pr "  unsigned int %s;\n" n
+    | UInt n | UIntPtr n -> pr "  unsigned int %s;\n" n
     | UInt32 n ->
        pr "  uint32_t %s_u32;\n" n;
        pr "  unsigned int %s; /* really uint32_t */\n" n
@@ -349,7 +368,7 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | SockAddrAndLen (n, _) -> pr " \"O\""
     | String n -> pr " \"s\""
     | StringList n -> pr " \"O\""
-    | UInt n -> pr " \"I\""
+    | UInt n | UIntPtr n -> pr " \"I\""
     | UInt32 n -> pr " \"I\""
     | UInt64 n -> pr " \"K\""
   ) args;
@@ -375,7 +394,7 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | SockAddrAndLen (n, _) -> pr ", &%s" n
     | String n -> pr ", &%s" n
     | StringList n -> pr ", &py_%s" n
-    | UInt n -> pr ", &%s" n
+    | UInt n | UIntPtr n -> pr ", &%s" n
     | UInt32 n -> pr ", &%s" n
     | UInt64 n -> pr ", &%s" n
   ) args;
@@ -440,13 +459,15 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
        pr "  %s = PyBytes_AS_STRING (py_%s);\n" n n;
        pr "  assert (%s != NULL);\n" n
     | SizeT n -> ()
-    | SockAddrAndLen _ ->
-       pr "  abort (); /* XXX SockAddrAndLen not implemented */\n";
+    | SockAddrAndLen (n, _) ->
+       pr "  if (nbd_internal_py_get_sockaddr (%s, &%s_sa, &%s_len) == -1)\n"
+         n n n;
+       pr "    goto out;\n"
     | String _ -> ()
     | StringList n ->
        pr "  %s = nbd_internal_py_get_string_list (py_%s);\n" n n;
        pr "  if (!%s) goto out;\n" n
-    | UInt _ -> ()
+    | UInt _ | UIntPtr _ -> ()
     | UInt32 n -> pr "  %s_u32 = %s;\n" n n
     | UInt64 n -> pr "  %s_u64 = %s;\n" n n
   ) args;
@@ -468,10 +489,10 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | Int64 n -> pr ", %s_i64" n
     | Path n -> pr ", %s" n
     | SizeT n -> pr ", (size_t)%s" n
-    | SockAddrAndLen (n, _) -> pr ", /* XXX */ (void *) %s, 0" n
+    | SockAddrAndLen (n, _) -> pr ", (struct sockaddr *) &%s_sa, %s_len" n n
     | String n -> pr ", %s" n
     | StringList n -> pr ", %s" n
-    | UInt n -> pr ", %s" n
+    | UInt n | UIntPtr n -> pr ", %s" n
     | UInt32 n -> pr ", %s_u32" n
     | UInt64 n -> pr ", %s_u64" n
   ) args;
@@ -520,6 +541,7 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | String _
     | StringList _
     | UInt _
+    | UIntPtr _
     | UInt32 _
     | UInt64 _ -> ()
   ) args;
@@ -533,7 +555,7 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | RErr ->
        pr "  py_ret = Py_None;\n";
        pr "  Py_INCREF (py_ret);\n"
-    | RFd | RInt | REnum _ | RFlags _ | RSizeT | RUInt ->
+    | RFd | RInt | REnum _ | RFlags _ | RSizeT | RUInt | RUIntPtr ->
        pr "  py_ret = PyLong_FromLong (ret);\n"
     | RInt64 | RCookie ->
        pr "  py_ret = PyLong_FromLongLong (ret);\n"
@@ -547,7 +569,9 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
   List.iter (
     function
     | Bool _ -> ()
-    | BytesIn (n, _) -> pr "  PyBuffer_Release (&%s);\n" n
+    | BytesIn (n, _) ->
+       pr "  if (%s.obj)\n" n;
+       pr "    PyBuffer_Release (&%s);\n" n
     | BytesOut (n, _) -> pr "  free (%s);\n" n
     | BytesPersistIn _ | BytesPersistOut _ -> ()
     | Closure { cbname } ->
@@ -565,6 +589,7 @@ let print_python_binding name { args; optargs; ret; may_set_error } =
     | UInt _ -> ()
     | UInt32 _ -> ()
     | UInt64 _ -> ()
+    | UIntPtr _ -> ()
   ) args;
   List.iter (
     function
@@ -806,6 +831,7 @@ class NBD(object):
           | UInt n -> n, None, None
           | UInt32 n -> n, None, None
           | UInt64 n -> n, None, None
+          | UIntPtr n -> n, None, None
         ) args in
       let optargs =
         List.map (
