@@ -1,5 +1,5 @@
 /* NBD client library in userspace
- * Copyright (C) 2013-2019 Red Hat Inc.
+ * Copyright (C) 2013-2021 Red Hat Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -31,8 +31,14 @@
 #include <errno.h>
 #include <sys/types.h>
 
+#ifdef HAVE_GNUTLS
+#include <gnutls/gnutls.h>
+#endif
+
 #include <libnbd.h>
 
+#include "../tests/requires.h"
+
 #define SIZE (1024*1024)
 
 #if CERTS || PSK
@@ -42,36 +48,64 @@
 #endif
 #endif
 
+#ifdef NEEDS_TMPFILE
+#define TMPFILE tmp
+static char tmp[] = "/tmp/nbdXXXXXX";
+
+static void
+unlink_tmpfile (void)
+{
+  unlink (TMPFILE);
+}
+#endif /* NEEDS_TMPFILE */
+
+#ifdef HAVE_GNUTLS
+#if TLS
+static void
+tls_log (int level, const char *msg)
+{
+  /* Messages from GnuTLS are always \n-terminated. */
+  fprintf (stderr, "gnutls[%d]: %s", level, msg);
+}
+#endif
+#endif
+
 int
 main (int argc, char *argv[])
 {
   struct nbd_handle *nbd;
-  char tmpfile[] = "/tmp/nbdXXXXXX";
   char *args[] = { SERVER, SERVER_PARAMS, NULL };
-  int fd;
   int64_t actual_size;
   char buf[512];
-  int r = -1;
+  size_t i;
+
+  /* Check requirements or skip the test. */
+#ifdef REQUIRES
+  REQUIRES
+#endif
 
   /* Create a large sparse temporary file. */
-  fd = mkstemp (tmpfile);
+#ifdef NEEDS_TMPFILE
+  int fd = mkstemp (TMPFILE);
   if (fd == -1 ||
       ftruncate (fd, SIZE) == -1 ||
       close (fd) == -1) {
-    perror (tmpfile);
-    goto out;
+    perror (TMPFILE);
+    exit (EXIT_FAILURE);
   }
+  atexit (unlink_tmpfile);
+#endif
 
   nbd = nbd_create ();
   if (nbd == NULL) {
     fprintf (stderr, "%s\n", nbd_get_error ());
-    goto out;
+    exit (EXIT_FAILURE);
   }
 
 #ifdef EXPORT_NAME
   if (nbd_set_export_name (nbd, EXPORT_NAME) == -1) {
     fprintf (stderr, "%s\n", nbd_get_error ());
-    goto out;
+    exit (EXIT_FAILURE);
   }
 #endif
 
@@ -80,6 +114,23 @@ main (int argc, char *argv[])
     fprintf (stderr, "skip: compiled without TLS support\n");
     exit (77);
   }
+
+#ifdef HAVE_GNUTLS
+  /* This is kind of ugly but GnuTLS only allows us to set these
+   * globally (so they are not appropriate for libnbd).
+   *
+   * Also by default GnuTLS throws away log messages even if you
+   * called gnutls_global_set_log_level.  It doesn't install the
+   * default log handler unless you set $GNUTLS_DEBUG_LEVEL.  So we
+   * need to have our own log handler.
+   *
+   * Also the log levels are quite random.  Level 2 doesn't show the
+   * negotiated cyphersuite, but level 3+ shows excessive detail.
+   */
+  gnutls_global_set_log_level (2);
+  gnutls_global_set_log_function (tls_log);
+#endif
+
   if (nbd_set_tls (nbd, TLS_MODE) == -1) {
     fprintf (stderr, "%s\n", nbd_get_error ());
     exit (EXIT_FAILURE);
@@ -102,6 +153,12 @@ main (int argc, char *argv[])
   }
 #endif
 
+  /* Print the server parameters. */
+  fprintf (stderr, "server: %s", args[0]);
+  for (i = 1; args[i] != NULL; ++i)
+    fprintf (stderr, " %s", args[i]);
+  fprintf (stderr, "\n");
+
   /* Start the server. */
 #if SOCKET_ACTIVATION
 #define NBD_CONNECT nbd_connect_systemd_socket_activation
@@ -110,7 +167,7 @@ main (int argc, char *argv[])
 #endif
   if (NBD_CONNECT (nbd, args) == -1) {
     fprintf (stderr, "%s\n", nbd_get_error ());
-    goto out;
+    exit (EXIT_FAILURE);
   }
 
 #if TLS
@@ -119,7 +176,7 @@ main (int argc, char *argv[])
       fprintf (stderr,
                "%s: TLS required, but not negotiated on the connection\n",
                argv[0]);
-      goto out;
+      exit (EXIT_FAILURE);
     }
   }
   else if (TLS_MODE == LIBNBD_TLS_ALLOW) {
@@ -128,14 +185,14 @@ main (int argc, char *argv[])
       fprintf (stderr,
                "%s: TLS disabled, but connection didn't fall back to plaintext\n",
                argv[0]);
-      goto out;
+      exit (EXIT_FAILURE);
     }
 #else // !TLS_FALLBACK
     if (nbd_get_tls_negotiated (nbd) != 1) {
       fprintf (stderr,
                "%s: TLS allowed, but not negotiated on the connection\n",
                argv[0]);
-      goto out;
+      exit (EXIT_FAILURE);
     }
 #endif
   }
@@ -144,17 +201,17 @@ main (int argc, char *argv[])
   actual_size = nbd_get_size (nbd);
   if (actual_size == -1) {
     fprintf (stderr, "%s\n", nbd_get_error ());
-    goto out;
+    exit (EXIT_FAILURE);
   }
   if (actual_size != SIZE) {
     fprintf (stderr, "%s: actual size %" PRIi64 " <> expected size %d",
              argv[0], actual_size, SIZE);
-    goto out;
+    exit (EXIT_FAILURE);
   }
 
   if (nbd_pread (nbd, buf, sizeof buf, 0, 0) == -1) {
     fprintf (stderr, "%s\n", nbd_get_error ());
-    goto out;
+    exit (EXIT_FAILURE);
   }
 
   /* XXX In future test more operations here. */
@@ -169,15 +226,11 @@ main (int argc, char *argv[])
    */
   if (nbd_shutdown (nbd, 0) == -1) {
     fprintf (stderr, "%s\n", nbd_get_error ());
-    goto out;
+    exit (EXIT_FAILURE);
   }
 #endif
 
   nbd_close (nbd);
 
-  r = 0;
- out:
-  unlink (tmpfile);
-
-  exit (r == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+  exit (EXIT_SUCCESS);
 }
